import subprocess
import shutil
import os
from pathlib import Path
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image

from espresso.utils import absolute_path as path, silent_remove

#--------------------------------------------------------------------------------------------

# This library is a python interface to Nick Rawlinson's 2D Fast Marching Fortran package fm2dss.f90
#
# M. Sambridge 
# July 2022
#--------------------------------------------------------------------------------------------

#  Definitions for waveTracker followed by those from Andrew Valentine's rayTracer.py package

# routines to write data files for fmst
from scipy.sparse import csr_matrix
from scipy.interpolate import RectBivariateSpline
import subprocess

class fmmResult(object):
    # Class variable
    
    def __init__(self):
        self.name = 'Travel time field object'
    def setTimes(self,t):
        self.ttimes = t.copy()
    def setPaths(self,p):
        self.paths = p.copy()
    def setFrechet(self,f):
        self.frechet = f.copy()
    def setTfield(self,w,source):
        self.tfield = w.copy()
        self.tfieldsource = source
        
class gridModel(object):
    def __init__(self,velocities,extent=(0,1,0,1),dicex=8,dicey=8):
        self.nx,self.ny = velocities.shape
        self.velocities=velocities
        self.xmin,self.xmax,self.ymin,self.ymax = extent
        self.xx = np.linspace(self.xmin,self.xmax,self.nx+1)
        self.yy = np.linspace(self.ymin,self.ymax,self.ny+1)
        self.dicex = dicex
        self.dicey = dicey
        self.extent = extent
    def getVelocity(self):
        return self.velocities.copy()
    def getSlowness(self):
        return 1./self.velocities # No copy needed as operation must return copy
    def setVelocity(self,v):
        assert self.velocities.shape == v.shape
        self.velocities = v.copy()
    def setSlowness(self,s):
        assert self.velocities.shape == s.shape
        self.velocities = 1./s
        

    def wavefront_tracker(self,recs,srcs,wdir='wavetracker/fmst',
            verbose=False,paths=False,frechet=False,times=True,tfieldsource=-1,
            tfilename="rtravel.out",ffilename="frechet.out",wfilename="travelt.out",rfilename="raypath.out",
            sourcegrid=1,sourcedice=5,sourcegridsize=10,
            earthradius=6371.0,schemeorder=1,nbsize=0.5,degrees=False,velocityderiv=False):

        v = self. velocities # velocity model node values

        recs = recs.reshape(-1, 2)
        srcs = srcs.reshape(-1, 2)
        if(tfieldsource+1 > len(srcs)): # source requested for travel time field does not exist
            print('Error: Travel time field corresponds to source:',tfieldsource,'\n',
                  '      but total number of sources is',len(srcs),
                  '\n       No travel time field will be calculated.\n')   
        
        # fmst expects input spatial co-ordinates in degrees and velocities in kms/s so we adjust (unless degrees=True)

        kms2deg = 180./(earthradius*np.pi)

        #write out input files for Fast Marching wavefront tracker fmst

        write_fm2dss_input(wdir,paths=paths,frechet=frechet,times=times,tfieldsource=tfieldsource+1,
                           dicex=self.dicex,dicey=self.dicey,sourcegrid=sourcegrid,sourcedice=sourcedice,
                           sourcegridsize=sourcegridsize,earthradius=earthradius,schemeorder=schemeorder,nbsize=nbsize,
                           tfilename=tfilename,ffilename=ffilename,wfilename=wfilename,rfilename=rfilename)  # write out control file

        write_rs(recs,srcs,wdir)    # write out sources and receiver files

        noncushion,nodemap = write_gridc(v,self.extent,wdir) # write data for input velocity model file gridc.txc
    
        write_otimes([[True]*len(recs)]*len(srcs),wdir) # write out rays to be calculated
    
        # run fmst wavefront tracker code from command line
        # see if the executable is there, otherwise do preparation for the executable
        out = run_fm2dss(wdir)
        # if out.returncode:      # re-compile if there's an error
        #     compile_fm2dss()
        #     out = run_fm2dss(wdir)
        if out.returncode:      # add permission if there's a further error
            print("Trying to fix now...")
            try:
                import stat
                exe_file = Path(wdir + "/build/fm2dss.o")
                st = os.stat(exe_file)
                os.chmod(exe_file, st.st_mode | stat.S_IEXEC)
                print("Execute permission given to fm2dss.o.")
            except:
                print("Failed to fix. Check error message above.")
            out = run_fm2dss(wdir)
        if(verbose): print(' Message from fmm2dss:',out.stdout)
        if(out.returncode != 0):
            print(' The process returned with errorcode:',out.returncode)
            print(' stdout: \n',out.stdout)
            print(' stderr: \n',out.stderr)
            return
        
        # collect results
        if(times):
            ttimes = read_fmst_ttimes(wdir+'/'+tfilename)
            if(not degrees): ttimes*= kms2deg # adjust travel times because inputs are not in degrees
    
        if(paths): 
            raypaths = read_fmst_raypaths(wdir+'/'+rfilename)
    
        if(frechet):
            frechetvals = read_fmst_frechet(wdir+'/'+ffilename,noncushion,nodemap)
            if(not degrees): frechetvals*= kms2deg # adjust travel times because inputs are not in degrees
            if(not velocityderiv): 
                x2 = -(v*v).reshape(-1)
                frechetvals = frechetvals.multiply(x2)

        if(tfieldsource>=0):
            tfieldvals = read_fmst_wave(wdir+'/'+wfilename)
            if(not degrees): tfieldvals*= kms2deg # adjust travel times because inputs are not in degrees
        
    #   build class object to return
        result = fmmResult()
    
        if(times): result.setTimes(ttimes)
        if(paths): result.setPaths(raypaths)
        if(frechet): result.setFrechet(frechetvals)
        if(tfieldsource >-1): result.setTfield(tfieldvals,tfieldsource) # set traveltime field and source id
        return result

def write_fm2dss_input(wdir,
    dicex=8,dicey=8,sourcegrid=1,sourcedice=5,sourcegridsize=10,
    earthradius=6371.0,schemeorder=1,nbsize=0.5,times=True,frechet=False,tfieldsource=0,paths=False,
    tfilename="rtravel.out",ffilename="frechet.out",wfilename="travelt.out",rfilename="raypath.out"):

    #dicex,dicey=8,8              # p-cell: dicing of propagation grid cell wrt to a velocity grid cell
    #sourcegrid=1                 # decide to use a finer propagation grid about each source (y=1,n=0)
    #sourcedice                   # s-cell: number of source cells per propagation grid p-cell;
    # ourcegridsize               # radius of source box in p grid cells 
    #earthradius                  # radius used to convert co-lat and long to kms (spherical Earth approx)
    #schemeorder                  # First-order(0) or mixed-order(1)
    #nbsize                       # narrow band size
    #lttimes = 1                  # bool to calculate travel times (y=1,n=0)
    #tfilename = "rtravel.out"    # filename for output travel times
    #lfrechet = 1                 # bool to calculate Frechet derivatives (0=no,1=yes)
    #ffilename                    # filename for output of  Frechet derivatives
    #tfieldsource                    # Id to calculate travel time field (0=no,>0=source id)
    #wfilename                    # filename for output of travel times
    #lpaths                       # Write out raypaths (<0=all,0=no,>0=source id)
    #rfilename                    # filename for output of raypaths
    lpaths = 0
    if(paths): lpaths=-1
    lttimes=0
    if(times): lttimes = 1
    lfrechet=0
    if(frechet): lfrechet = 1
    
    blank = "c"*70+"\n"
    f = open(wdir+"/fm2dss.in", "w+")
    f.write(blank+"c INPUT PARAMETERS\n"+blank)
    l1 = "sources.dat                    c: File containing source positions\n"
    l2 = "receivers.dat                  c: File containing receiver positions\n"
    l3 = "otimes.dat                     c: File containing source-receiver associations\n"
    l4 = "gridc.vtx                      c: File containing velocity grid information\n"
    f.write(l1+l2+l3+l4)
    string = "{}     {}                        c: Grid dicing in latitude and longitude\n".format(dicex,dicey)
    f.write(string)
    string = "{}                              c: Apply source grid refinement? (0=no,1=yes)\n".format(sourcegrid)
    f.write(string)
    string = "{}     {}                       c: Dicing level and extent of refined grid\n".format(sourcedice,sourcegridsize)
    f.write(string)
    string = "{}                         c: Earth radius in km\n".format(earthradius)
    f.write(string)
    string = "{}                              c: Use first-order(0) or mixed-order(1) scheme\n".format(schemeorder)
    f.write(string)
    string = "{}                            c: Narrow band size (0-1) as fraction of nnx*nnz\n".format(nbsize)
    f.write(string)
    f.write(blank+"c OUTPUT FILES\n"+blank)
    string = "{}                              c: find source-receiver traveltimes (0=no,1=yes)\n".format(lttimes)
    f.write(string)
    f.write(tfilename.ljust(30)+"c: Name of file containing source-receiver traveltimes\n")
    string = "{}                              c: Calculate Frechet derivatives (0=no,1=yes)\n".format(lfrechet)
    f.write(string)
    f.write(ffilename.ljust(30)+"c: Name of file containing Frechet derivatives\n")
    string = "{}                              c: Write traveltime field to file? (0=no,>0=source id)\n".format(tfieldsource)
    f.write(string)
    f.write(wfilename.ljust(30)+"c: Name of file containing traveltime field\n")
    string = "{}                             c: Write out raypaths (<0=all,0=no,>0=source id)\n".format(lpaths)
    f.write(string)
    f.write(rfilename.ljust(30)+"c: Name of file containing raypath geometry\n")

    f.close()
    
def write_rs(recs,srcs,wdir): # write sources and receivers in format for fmst
    f = open(wdir+"/receivers.dat", "w+")
    f.write(" {} \n".format(len(recs)))
    for i in range(len(recs)):
        f.write(" {} {} \n".format(recs[i,1],recs[i,0]))
    f.close()
    f = open(wdir+"/sources.dat", "w+")
    f.write(" {} \n".format(len(srcs)))
    for i in range(len(srcs)):
        f.write(" {} {} \n".format(srcs[i,1],srcs[i,0]))
    f.close()

def write_gridc(v,extent,wdir): # write data for input velocity model file gridc.txc
    #
    # here extent[3],extent[2] is N-S range of grid nodes
    #      extent[0],extent[1] is W-E range of grid nodes
    nx,ny = v.shape
    dlat, dlong = (extent[3]-extent[2])/(ny-1),(extent[1]-extent[0])/(nx-1) # grid node spacing in lat and long

    # gridc.vtx requires a single cushion layer of nodes surrounding the velocty model
    # build velocity model with cushion velocities

    noncushion = np.zeros((nx+2,ny+2),dtype=bool) # bool array to identify cushion and non cushion nodes
    noncushion[1:nx+1,1:ny+1] = True
    
    # mapping from cushion indices to non cushion indices
    nodemap = np.zeros((nx+2,ny+2),dtype=int)
    nodemap[1:nx+1,1:ny+1] = np.array(range((nx*ny))).reshape((nx,ny))
    nodemap=nodemap[:,::-1]

    # build velocity nodes
    # additional boundary layer of velocities are duplicates of the nearest actual velocity value.
    vc = np.ones((nx+2,ny+2))
    vc[1:nx+1,1:ny+1] = v
    vc[1:nx+1,0] = v[:,0]   # add velocities in the cushion boundary layer
    vc[1:nx+1,-1] = v[:,-1] # add velocities in the cushion boundary layer
    vc[0,1:ny+1] = v[0,:]   # add velocities in the cushion boundary layer
    vc[-1,1:ny+1] = v[-1,:] # add velocities in the cushion boundary layer
    vc[0,0],vc[0,-1],vc[-1,0],vc[-1,-1] = v[0,0],v[0,-1],v[-1,0],v[-1,-1]
    vc=vc[:,::-1]

    # write out gridc.vtx file
    dummy = 1.0 # uncertainty in velocity grid parameter (not required here)
    f = open(wdir+"/gridc.vtx", "w+")
    f.write(" {} {} \n".format(ny,nx)) # Number of grid nodes in latitude and longitude 
    f.write(" {} {} \n".format(extent[3],extent[0])) # origin of computational grid
    f.write(" {} {} \n\n".format(dlat, dlong)) # grid node spacing in lat and long
 
    for i in range(nx+2):
        for j in range(ny+2):
            f.write(" {} {} \n".format(vc[i,j],dummy))
        f.write("\n")
    f.close()
    
    return noncushion,nodemap.flatten()

def write_otimes(pathsTF,wdir): # write out paths that are True in 2D list
    f = open(wdir+"/otimes.dat", "w+")
    ns = len(pathsTF)
    nr = len(pathsTF[0])
    for i in range(ns):
        for j in range(nr):
            if(pathsTF[i][j]): 
                f.write(" {} \n".format(1))
            else:
                f.write(" {} \n".format(0))
    f.close()   

def read_fmst_raypaths(filename): # read ray paths from output of fmst
    with open(filename, 'r') as f:
        lines = f.readlines()
        columns = lines[0].split()
        nrays = int(columns[0])
        paths = []
        k=1
        for i in range(nrays):
            lat = []
            long = []
            npts = float(lines[k])
            columns = lines[k].split()
            npts = int(columns[0])
            k+=1
            for j in range(npts):
                columns = lines[k].split()
                lat.append(float(columns[0]))
                long.append(float(columns[1]))
                k+=1
            path = np.array([long,lat]).T
            paths.append(path) 
    f.close()
    return paths 

def read_fmst_ttimes(filename): # read travel times from output of fmst
    with open(filename, 'r') as f:
        lines = f.readlines()
        ids = []
        ttimes = []
        for line in lines:
            columns = line.split()
            ids.append(int(columns[0]))
            ttimes.append(float(columns[1]))
    f.close()
    return np.asarray(ttimes)

def read_fmst_frechet(filename,noncushion,nodemap): # read ray paths from output of fmst
    nx,ny = noncushion.shape
    nx,ny=nx-2,ny-2
    mused = noncushion.flatten()
    frechet = []
    # the file frechet.out is written by a Fortran routine with array indices starting from 1 and
    # includes frechet derivatives for cushion entries that we do not want
    ray = 0 # ray index
    with open(filename, 'r') as f:
        lines = f.readlines()
        k=0
        vals = [] # set up csr_matrix components
        cols = []
        rows = []
        while k < len(lines):
            columns = lines[k].split()
            nparams = int(columns[0])
            k+=1
            kk = 0
            while kk < nparams:
                columns = lines[k].split()
                mod_id = int(columns[0])-1 # input indices start from 1 so adjust.
                if(mused[mod_id]): # ignore model entries corresponding to cushion nodes
                    # map node indices including cushion to original node indices without cushion
                    rows.append(ray)
                    cols.append(nodemap[mod_id])
                    vals.append(float(columns[1]))
                kk+=1
                k+=1
            ray+=1
    f.close()
    
    # convert to csr matrix
    frechet = csr_matrix((vals, (rows, cols)), shape=(ray, nx*ny))
    return frechet

def read_fmst_wave(filename):
    with open(filename, 'r') as f:
        # The grid definition used by fm2sds.f90 here is identical to the v grid specified by extent
        # with origin at the minimum colat and long, equivalent top left on 2D image with z down the page and x to right.
        lines = f.readlines()
        columns = lines[0].split()
        goxd,gozd = float(columns[0]),float(columns[1]) # Origin of grid (theta,phi) degress or long,lat
        columns = lines[1].split()
        nnx,nnz = int(columns[0]),int(columns[1]) # Number of nodes of grid in x and z or long,lat
        columns = lines[2].split()
        dnxd,dnzd = float(columns[0]),float(columns[1]) # cell sizes in x and z (theta,phi) degress or long,lat
        tfield = np.zeros((nnz,nnx))
        k = 3
        for i in range(nnz):
            for j in range(nnx):
                columns = lines[k].split()
                tfield[i,j] = float(columns[0])
                k+=1
    f.close()
    tfield = tfield[:,::-1]
    return tfield

def displayModel(model,paths=None,extent=(0,1,0,1),clim=None,cmap=None,
                 figsize=(6,6),title=None,line=1.0,cline='k',alpha=1.0,wfront=None,cwfront='k',
                 diced=True,dicex=8,dicey=8,cbarshrink=0.6,**wkwargs):
    fig = plt.figure(figsize=figsize)
    if cmap is None: cmap = plt.cm.RdBu

    # if diced option plot the actual B-spline interpolated velocity used by fmst program
    
    plotmodel = model
    if(diced):
        plotmodel = dicedgrid(model,extent=extent,dicex=dicex,dicey=dicey) 
    
    plt.imshow(plotmodel.T,origin='lower',extent=extent,cmap=cmap)


    if paths is not None:
        if(isinstance(paths, np.ndarray) and paths.shape[1] == 4): # we have paths from xrt.tracer so adjust
            paths = changepathsformat(paths)

        for p in paths:
            plt.plot(p[:,0],p[:,1],cline,lw=line,alpha=alpha)
    if clim is not None: plt.clim(clim)
    
    if title is not None: plt.title(title)
    
    if(wfront is not None):
        nx,ny = wfront.shape
        X, Y = np.meshgrid(np.linspace(extent[0],extent[1],nx), np.linspace(extent[2],extent[3],ny))
        plt.contour(X, Y, wfront.T, **wkwargs)  # Negative contours default to dashed.
    
    if(wfront is None): plt.colorbar(shrink=cbarshrink)

    # plt.show()
    return fig

def dicedgrid(v,extent=[0.,1.,0.,1.],dicex=8,dicey=8):    
    nx,ny = v.shape
    x = np.linspace(extent[0], extent[1],nx)
    y = np.linspace(extent[2], extent[3],ny)
    kx,ky=3,3
    if(nx <= 3): kx = nx-1 # reduce order of B-spline if we have too few velocity nodes
    if(ny <= 3): ky = ny-1 
    rect = RectBivariateSpline(x, y, v,kx=kx,ky=ky)
    xx = np.linspace(extent[0], extent[1],dicex*nx)
    yy = np.linspace(extent[2], extent[3],dicey*ny)
    X,Y = np.meshgrid(xx,yy,indexing='ij')
    vinterp = rect.ev(X,Y)
    return vinterp

def changepathsformat(paths):
    p = np.zeros((len(paths),2,2))
    for i in range(len(paths)):
        p[i,0,0] = paths[i,0]
        p[i,0,1] = paths[i,1]
        p[i,1,0] = paths[i,2]
        p[i,1,1] = paths[i,3]
    return p

#--------------------------------------------------------------------------------------------
def norm(x):
    return np.sqrt(x.dot(x))
def normalise(x):
    return x/norm(x)
def pngToModel(pngfile,nx,ny,bg=1.,sc=1.):
    png = Image.open(pngfile)
    png.load()

    model = sc*(bg+np.asarray(png.convert('L').resize((nx,ny)).transpose(Image.ROTATE_270))/255.)
    return model

def generateSurfacePoints(nPerSide,extent=(0,1,0,1),surface=[True,True,True,True],addCorners=True):
    out = []
    if surface[0]:
        out+=[[extent[0],x] for x in np.linspace(extent[2],extent[3],nPerSide+2)[1:nPerSide+1]]
    if surface[1]:
        out+=[[extent[1],x] for x in np.linspace(extent[2],extent[3],nPerSide+2)[1:nPerSide+1]]
    if surface[2]:
        out+=[[x,extent[2]] for x in np.linspace(extent[0],extent[1],nPerSide+2)[1:nPerSide+1]]
    if surface[3]:
        out+=[[x,extent[3]] for x in np.linspace(extent[0],extent[1],nPerSide+2)[1:nPerSide+1]]
    if addCorners:
        if surface[0] or surface[2]:
            out+=[[extent[0],extent[2]]]
        if surface[0] or surface[3]:
            out+=[[extent[0],extent[3]]]
        if surface[1] or surface[2]:
            out+=[[extent[1],extent[2]]]
        if surface[1] or surface[3]:
            out+=[[extent[1],extent[3]]]
    return np.array(out)

def run_fm2dss(wdir):
    command = "./build/fm2dss.o"
    return subprocess.run(command,stdout=subprocess.PIPE, text=True,shell=True,cwd=wdir)

def compile_fm2dss():
    # https://github.com/inlab-geo/espresso/blob/main/espresso_machine/build_package/validate.py#L170
    build_dir = path("./build")
    res1 = subprocess.call(["cmake", ".."], cwd=build_dir)
    if res1:
        raise ChildProcessError(f"`cmake .` failed in {build_dir}")
    res2 = subprocess.call(["make"], cwd=build_dir)
    if res2:
        raise ChildProcessError(f"`make` failed in {build_dir}")

def clean_fm2dss():
    shutil.rmtree(path("build"))



#-----------------------------------------------------------------------
# This script is an interface between Receiver function routines and python
# 
# Most parameters relating to the Receiver Function are set within the Fortran routine.
#
# The fortran interface was generated with f2py using the call
# > f2py -m RF -c RF.f90 RFsubs/*.f*
#
# Fortran routines for Receiver function calculation by T. Shibutani
#
# M. Sambridge, 
# RSES, Oct., 2017.
#
#-----------------------------------------------------------------------
#import sys
try:
    from .build import rfc as rfm # if file is only to be used as library
except:
    from build import rfc as rfm # if file is to be executed
import numpy as np
#matplotlib.use('Qt4Agg')
#import matplotlib
from matplotlib import pyplot as plt
##################################################################################
def rfcalc(model,sn=0.0,mtype=0,fs=25.0,gauss_a=2.5,water_c=0.0001,angle=35.0,time_shift=5.0,ndatar=626,v60=8.043,seed=1): # Calculate Data covariance matrix for evaluation of waveform fit
    #print ('sn ',sn)
    if(sn==0.0):
        a,b = rfm.rfcalc_nonoise(model,mtype,fs,gauss_a,water_c,angle,time_shift,ndatar,v60)
    else:
        a,b = rfm.rfcalc_noise(model,mtype,sn,fs,gauss_a,water_c,angle,time_shift,ndatar,v60,seed)
    return a,b
##################################################################################
def InvDataCov(width,Ar,ndata): # Calculate Data covariance matrix for evaluation of waveform fit
    sigsq = width**2
    Arsq = Ar*Ar # waveform noise variance    
    Cd = np.zeros((ndata,ndata))
    for k1 in range(ndata):
        for k2 in range(ndata):
            Cd[k1][k2] = np.exp(-0.5*(k1-k2)**2/sigsq)
    U, s, V = np.linalg.svd(Cd)
    pmax = [ n for n,a in enumerate(s) if a<0.000001 ][0]
    Sinv = np.diag(1.0/s[:pmax])
    Cdinv = np.dot(V.T[:,:pmax], np.dot(Sinv, U.T[:pmax,:]))
    return Cdinv/Arsq

##################################################################################
def InvDataCovSub(width,Ar,ndata,ind): # Calculate Data sub-covariance matrix for evaluation of waveform fit
    sigsq = width**2
    Arsq = Ar*Ar # waveform noise variance    
    Cd = np.zeros((ndata,ndata))
    for k1 in range(ndata):
        for k2 in range(ndata):
            Cd[k1][k2] = np.exp(-0.5*(k1-k2)**2/sigsq)
    Cdused = Cd[ind,:][:,ind]
    U, s, V = np.linalg.svd(Cdused)
    pmax = [ n for n,a in enumerate(s) if a>0.000001 ][-1]+1
    Sinv = np.diag(1.0/s[:pmax])
    Cdinv = np.dot(V.T[:,:pmax], np.dot(Sinv, U.T[:pmax,:]))
    return Cdinv/Arsq        
##################################################################################
def plot_misfit_profile(x,misfit,xtrue,iparam): # Plot calculated and observed RF waveform 
    fig, ax = plt.subplots()
    
    if(iparam[1]==1): strx = "Layer "+repr(iparam[0]+1)+" velocity (km/s)" # plot axis labels
    if(iparam[1]==0): strx = "Layer "+repr(iparam[0]+1)+" node depth (km)"

    mismin = np.min(misfit)
    mismax = np.max(misfit)
    xt = [xtrue,xtrue]
    yt = [mismin,mismin+0.1*(mismax-mismin)]        
    ax.plot(x, misfit, 'k-')
    ax.plot(xt,yt,'r-',label='True')
    if(iparam[1] == 0):
        ax.set_xlabel('Node Depth (km)') 
    else:
        ax.set_xlabel('Velocity (km)/s')
    ax.set_xlabel(strx)
    #ax.set_ylabel('-Log L(m)')
    #ax.set_title("-Log-Likelihood through reference model")
    ax.set_ylabel('Misfit')
    ax.set_title("Waveform misfit profile")
    ax.grid(True)
    xt2 = [x[np.argmin(misfit)],x[np.argmin(misfit)]] 
    ax.plot(xt2,yt,'c-',label='Minimum') 
    plt.title("Misfit profile along single axis")
    plt.legend()
    plt.show()
##################################################################################
def l2mod(model,vmin=2.4,vmax=4.7,dmin=0.0,dmax=60.0): # Transform layer thickness representation to (depth vel) plot format
    px = np.zeros([2*len(model)])
    py = np.zeros([2*len(model)])
    py[1::2],py[2::2],px[0::2],px[1::2] = list(np.cumsum(model.T[0])),list(np.cumsum(model.T[0]))[:-1],model.T[1],model.T[1]
    py[-1] = dmax
    return px,py
##################################################################################
def v2mod(model,vmin=2.4,vmax=4.7,dmin=0.0,dmax=60.0): # Transform Voronoi nucleus representation to (depth vel) plot format
    a,b,c,d,e = rfm.voro2mod(model)
    px = np.zeros([2*len(d)])
    py = np.zeros([2*len(d)])
    a,b,c,d,e = rfm.voro2mod(model)
    py[1::2],py[2::2],px[0::2],px[1::2] = list(np.cumsum(a)),list(np.cumsum(a))[:-1],b[:],b[:]
    py[-1] = dmax
    return np.cumsum(a),b,c,d,e,px,py
##################################################################################
def d2mod(model,vmin=2.4,vmax=4.7,dmin=0.0,dmax=60.0): # Transform depth representation to (depth vel) plot format
    px = np.zeros([2*len(model)])
    py = np.zeros([2*len(model)])
    py[1::2],py[2::2],px[0::2],px[1::2] = list(model.T[0]),list(model.T[0])[:-1],model.T[1],model.T[1]
    py[-1] = dmax
    return px,py
##################################################################################
def plot_RFs(time1,RFo,time2,RFp,string="Observed and predicted receiver functions"):
    fig, ax = plt.subplots()
    plt.title(string)
    plt.xlabel("Time (s)")
    plt.ylabel("Amplitude")
    plt.grid(True)
    plt.plot(time1, RFo, 'k-', label='Observed')
    plt.plot(time2, RFp, 'r-', label='Predicted')
    plt.legend()
    plt.show()
    plt.savefig('RF_plot.pdf',format='PDF')
##################################################################################
def plotRFm(velmod,time1,RFo,time2,RFp,mtype = 0,vmin=2.4,vmax=4.7,dmin=0.0,dmax=60.0,string="Observed and predicted receiver functions"): # plot velocity model and receiver function
    f, (a0, a1) = plt.subplots(1,2, figsize=(12,4), gridspec_kw = {'width_ratios':[1, 3]})

    a1.set_title(string)
    a1.set_xlabel("Time (s)")
    a1.set_ylabel("Amplitude")
    #if((k==3) & (j==0)):a1.set_ylim(-0.2,0.6)
    #if((k==3) & (j==1)):a1.set_ylim(-0.12,0.4)
    a1.grid(True)
    a1.plot(time1, RFo, '-', color="grey",label='Observed')
    #a1.plot(time2[:626], RFp[:626], 'b-', label='Predicted')
    a1.plot(time2, RFp, 'b-', label='Predicted')
    a1.legend()

    if(mtype ==0):d, beta, vpvs, qa, qb, pv, pd = v2mod(velmod)  # Convert velocity model from Voronoi format to plot format
    if(mtype ==1):pv, pd = l2mod(velmod)  # Convert velocity model from Layer format to plot format
    if(mtype ==2):pv, pd = d2mod(velmod)  # Convert velocity model from Layer format to plot format
    
    a0.set_title(" Velocity model")                   # Plot velocity model with Receiver function
    a0.set_xlabel('Vs (km/s)')
    a0.set_ylabel('Depth (km)')
    a0.plot(pv,pd,'g-')
    a0.set_xlim(vmin,vmax)
    a0.set_ylim(dmin,dmax)
    a0.invert_yaxis()
    if(mtype==0): a0.plot(velmod[:,1],velmod[:,0],ls="",marker="o",markerfacecolor='none', markeredgecolor='k')

    #plt.tight_layout()
    plt.show()
    plt.savefig('RF_mod_plot.pdf',format='PDF')
    return pv,pd
##################################################################################
if __name__ == "__main__":
    #print(sys.path)
    velmod0 = np.zeros([7,3]) # Set up velocity reference model
    velmod0[0] = [8.370596, 3.249075, 1.7]
    velmod0[1] = [17.23163, 3.001270, 1.7]
    velmod0[2] = [1.9126695E-02, 2.509443, 1.7]
    velmod0[3] = [19.78145, 3.562691, 1.7]
    velmod0[4] = [41.73066, 4.225965, 1.7]
    velmod0[5] = [14.35261, 2.963322, 1.7]
    velmod0[6] = [49.92358, 4.586726, 1.7]
    velmod = np.copy(velmod0)

	# plot waveforms

    #time1, RFo = np.loadtxt("RF_obs.dat", unpack=True) # Read in observed Receiver function
	
    #time2, RFp = RF.rfcalc(velmod)                     # Calculate predicted Receiver function using velocity model
    time1, RFo = rfcalc(velmod) # Read in observed Receiver function
    time2, RFp = rfcalc(velmod,sn=0.5,seed=61254557)                     # Calculate predicted Receiver function using velocity model
    plt.title(" Observed and predicted receiver functions")
    plt.xlabel("Time (s)")
    plt.ylabel("Amplitude")
    plt.grid(True)
    plt.plot(time1, RFo, 'k-', label='Observed')
    plt.plot(time2[:626], RFp[:626], 'r-', label='Predicted')
    plt.legend()
    plt.show()
    plt.savefig('RF_plot.pdf',format='PDF')

	# Set up data Covariance matrix
	# 2.5, 0.01 = correlation half-width and amplitude of Gaussian noise
    ndata = len(RFp)
    Cdinv = InvDataCov(2.5,0.01,ndata) 

# Calculate waveform misfit for reference model

    res = RFo-RFp
    mref = 	np.dot(res,np.transpose(np.dot(Cdinv, res)))/2.0
    print (' Waveform misfit of reference model',mref)



import _utils


def pytest_addoption(parser):
    parser.addoption("--post", action="store_true", default=False)
    parser.addoption(
        "--timeout",
        dest="timeout",
        action="store",
        default=999,
        type=int,
        help="Specify the number of seconds as timeout limit for each attribute",
    )
    parser.addoption(
        "--contrib",
        "--contribution",
        dest="contribs",
        action="append",
        default=None,
        help="Specify which contribution to validate",
    )


def pytest_generate_tests(metafunc):
    if "contrib" in metafunc.fixturenames:
        specified = metafunc.config.getoption("contribs")
        problems = _utils.problems_to_run(specified)
        metafunc.parametrize("contrib", set(problems))



import errno
import os
import signal
import functools
import contextlib
import sys
import tqdm
import argparse
import warnings
import pathlib
import typing


# ------------------------------- constants -------------------------------------------
PKG_NAME = "espresso"
ROOT = str(pathlib.Path(__file__).resolve().parent.parent.parent)
CONTRIB_FOLDER = ROOT + "/contrib"
ACTIVE_LIST = CONTRIB_FOLDER + "/active_problems.txt"

DEFAULT_TIMEOUT = 60
DEFAULT_TIMEOUT_SHORT = 1


# ------------------------------- argument parser -------------------------------------
def setup_parser():
    parser = argparse.ArgumentParser(
        description="Script to build Espresso, with/without pre/post-build validation"
    )
    parser.add_argument(
        "-v",
        "--checks",
        "--post",
        "--validate",
        dest="post",
        action="store_true",
        default=False,
        help="Run tests after building the package",
    )
    parser.add_argument(
        "--pre",
        dest="pre",
        action="store_true",
        default=False,
        help="Run tests before building the package",
    )
    parser.add_argument(
        "--timeout",
        "-t",
        dest="timeout",
        action="store",
        default=999,
        type=int,
        help="Specify the number of seconds as timeout limit for each attribute",
    )
    parser.add_argument(
        "--contrib",
        "-c",
        "--contribution",
        dest="contribs",
        action="append",
        default=None,
        help="Specify which contribution to validate",
    )
    parser.add_argument(
        "--file",
        "-f",
        dest="file",
        action="store",
        default=None,
        help="Specify which contributions to validate in a txt file",
    )
    parser.add_argument("--all", "-a", default=None, dest="all", action="store_true")
    parser.add_argument(
        "--no-install", default=False, dest="dont_install", action="store_true"
    )
    return parser


def args():
    args, unknown = setup_parser().parse_known_args()
    return args


def pre_build():
    _args = args()
    return _args.pre or (not _args.pre and not _args.post)


# ------------------------------- running timeout -------------------------------------
# Thanks to: https://stackoverflow.com/a/2282656
def timeout(seconds=10, error_message=os.strerror(errno.ETIME)):
    def decorator(func):
        def _handle_timeout(signum, frame):
            raise TimeoutError(error_message)

        if seconds is None:

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

        else:

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                signal.signal(signal.SIGALRM, _handle_timeout)
                signal.alarm(seconds)
                try:
                    result = func(*args, **kwargs)
                finally:
                    signal.alarm(0)
                return result

        return wrapper

    return decorator


# ------------------------------- hide console output ---------------------------------
# Thanks to:
# https://stackoverflow.com/a/25061573
# https://stackoverflow.com/a/72617364
@contextlib.contextmanager
def suppress_stdout():
    def tqdm_replacement(iterable_object, *args, **kwargs):
        return iterable_object

    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        tqdm_copy = tqdm.tqdm  # store it if you want to use it later
        tqdm.tqdm = tqdm_replacement
        try:
            yield
        finally:
            sys.stdout = old_stdout
            tqdm.tqdm = tqdm_copy


# ------------------------------- contribution processing -----------------------------
def problem_name_to_class(problem_name):  # e.g. "xray_tracer" -> "XrayTracer"
    return problem_name.title().replace("_", "")


def get_folder_content(folder_name) -> typing.Tuple[typing.List[str], typing.List[str]]:
    names = [name for name in os.listdir(folder_name)]
    paths = [f"{folder_name}/{name}" for name in names]
    return names, paths


def problems_specified_from_args():
    contribs_file = args().file
    if contribs_file is not None:
        with open(contribs_file, "r") as f:
            lines = f.readlines()
        contribs = []
        for line in lines:
            if line.startswith("#"):
                continue
            contribs.append(line.strip())
    else:
        contribs = args().contribs
    return contribs

def problems_to_run(problems_specified=None) -> typing.List[typing.Tuple[str, str]]:
    if problems_specified is None:
        problems_specified = problems_specified_from_args()
    all_problems = get_folder_content(CONTRIB_FOLDER)
    all_problems_zipped = list(zip(*all_problems))
    all_problems_zipped = [c for c in all_problems_zipped if "." not in c[0]]
    if problems_specified is None:
        return all_problems_zipped
    else:  # filter by specified problems list
        problems = [c for c in all_problems_zipped if c[0] in problems_specified]
        problems_not_in_folder = [
            c for c in problems_specified if c not in all_problems[0]
        ]
        if problems_not_in_folder:
            warnings.warn(
                "these examples are not detected in 'contrib' folder: "
                + ", ".join(problems_not_in_folder)
            )
        return problems

def problems_to_run_names_only(problems_specified=None) -> typing.List[str]:
    problems = problems_to_run(problems_specified)
    problem_names = [c[0] for c in problems]
    return problem_names



"""Test all / specified contributed problems, and this depends on:
1. results generated by: run_examples.py
2. criteria specified in: criteria.py
3. problems identified in: conftest.py

Usage: python test_examples.py [--post] [--contribution <example_name>] [--file <file_name>]
"""

import pathlib
import pytest
import json
import re

import report
import _utils


@pytest.fixture
def pre_build(request):
    return not request.config.getoption("--post")


# Note: `contrib` fixture is defined in conftest.py
def test_contrib(contrib, pre_build):
    _report = report.compliance_report([contrib[0]], pre_build)
    for _r in _report.values():  # always only one iteration
        report.pprint_compliance_report(_report)
        if isinstance(_r.api_compliance, Exception):
            raise _r.api_compliance
        assert _r.api_compliance, \
            "Not API-compliant. Check report above for details."


pass_pattern = r"\(([\w\s]+)\) is API-compliant"

def write_active_list():
    # read report
    with open(".report.json", "r") as f:
        report = json.load(f)
    # analyse report
    active_list = []
    for r in report["tests"]:
        match = re.search(pass_pattern, r["call"]["stdout"])
        if match:
            active_list.append(match.group(1))
    # write to file
    active_list.sort()
    with open(_utils.ACTIVE_LIST, "w") as f:
        f.write(f"# augo generated by {__file__}\n")
        f.write("\n".join(active_list))


def main():
    this_file = pathlib.Path(__file__)
    pytest.main([this_file, "--json-report", "--json-report-indent=2"])
    write_active_list()


if __name__ == "__main__":
    main()



r"""

EspressoProblem
---------------

The essential idea of Espresso is to have exactly same interface for different 
inversion test problems, so we list all the standard methods and attributes once in 
this API reference page. Check out documentation for :class:`EspressoProblem` for 
details.

.. autosummary::
    :toctree: generated/
    :nosignatures:

    EspressoProblem

List the problems
-----------------

To programatically see what problems are available in Espresso, these functions are 
helpful:

.. autofunction:: list_problems
.. autofunction:: list_problem_names

Utility functions
-----------------

Some utility functions are there to help contributors load data and calculate things.
Check out documentation for submodule :mod:`espresso.utils` for details.

.. autosummary::
    :toctree: generated/
    :nosignatures:
    
    espresso.utils

"""

from ._espresso_problem import EspressoProblem


from ._version import __version__


__all__ = [
    "EspressoProblem",
]

# from .example_name import ExampleName

# from .list_problems import list_problem_names, list_problems
# __all__ += list_problem_names()
# __all__ += ["list_problem_names", "list_problems"]

from ._magnetotelluric_1D import Magnetotelluric1D
from ._slug_test import SlugTest
from ._fm_wavefront_tracker import FmWavefrontTracker
from ._gravity_density import GravityDensity
from ._pumping_test import PumpingTest
from ._xray_tracer import XrayTracer
from ._receiver_function import ReceiverFunction
from ._simple_regression import SimpleRegression

from .list_problems import list_problem_names, list_problems
__all__ += list_problem_names()
__all__ += ['list_problem_names', 'list_problems']



import pathlib
import numpy as np
import matplotlib.pyplot as plt

from espresso import EspressoProblem
from espresso.exceptions import InvalidExampleError


LIB_DIR = pathlib.Path(__file__).resolve().parent / "lib"

class ReceiverFunction(EspressoProblem):
    """Forward simulation class
    """

    metadata = {
        "problem_title": "Receiver function",                # To be used in docs
        "problem_short_description": (
            "'Receiver functions' are a class of seismic data used to study "
            "discontinuities (layering) in the Earth's crust"
        ),    # 1-3 sentences

        "author_names": ["Malcolm Sambridge","Takuo Shibutani"],    # List of names e.g. author_names = ["Sally Smith", "Mark Brown"]

        "contact_name": "Malcolm Sambridge",         # Contact for contributor/maintainer of espresso example
        "contact_email": "Malcolm.Sambridge@anu.edu.au",

        "citations": [("Langston, C. A., Structure under Mount Rainer, Washington, inferred from teleseismic body waves, J. Geophys. Res., vol 84, 4749-4762, 1979.", 
                       "Shibutani, T., Kennett, B. and Sambridge, M., Genetic algorithm inversion for receiver functions with application to crust and uppermost mantle structure beneath Eastern Australia, Geophys. Res. Lett., 23 , No. 4, 1829-1832, 1996")], # Reference to publication(s) that describe this example. In most 
                                # cases there will only be a single entry in this list.
                                # List of (citation, doi) pairs e.g. 
                                # citations = [("Newton, I (1687). Philosophiae naturalis principia mathematica.", "")]
                                # If there are no citations, use empty list `[]`

        "linked_sites": [],  # List of (title, address) pairs for any websites that 
                                    # should be linked in the documentation, e.g.
                                    # linked_sites = [("Parent project on Github","https://github.com/user/repo"),
                                    #                 ("Data source"),"https://www.data.com") ]
                                    # If there are no links, use empty list `[]`
    }


    def __init__(self, example_number=1):
        super().__init__(example_number)

        from . import rf
        self.rf = rf

        # example initialisation
        self._ref_model_setup = np.array([[1,4.0,1.7],
                                        [3.5,4.3,1.7],
                                        [8.0,4.2,2.0],
                                        [20, 6,1.7],
                                        [45,6.2,1.7]])
        self._t, self._data = self.rf.rfcalc(self._ref_model_setup, sn=0.5)
        if example_number == 1:
            self._description = "Inverting depths of the 2nd and 3rd interfaces"
            self._good_model = np.array([8., 20.])
            self._null_model = np.array([10., 30.])
            self._interfaces = [2, 3]      # 1st and 2nd interfaces for inversion
            self._nmodel = len(self._interfaces)
        elif example_number == 2:
            self._description = "Inverting velocities of 5 layers"
            self._good_model = np.array([4., 4.3, 4.2, 6., 6.2])
            self._null_model = np.ones(5) * 4.5
            self._nmodel = self._ref_model_setup.shape[0]
        elif example_number == 3:
            self._description = "Inverting depths and velocities of 5 layers"
            self._good_model = self._ref_model_setup[:,:2].flatten()
            _null_depths = np.array([10,20,30,40,50])
            _null_velocities = np.ones(5) * 4.5
            _null_model = np.vstack((_null_depths, _null_velocities))
            self._null_model = _null_model.T.flatten()
            self._nmodel = len(self._null_model)
        else:
            raise InvalidExampleError

    @property
    def description(self):
        return self._description

    @property
    def model_size(self):
        return self._nmodel

    @property
    def data_size(self):
        return len(self._data)

    @property
    def good_model(self):
        return self._good_model

    @property
    def starting_model(self):
        return self._null_model

    @property
    def data(self):
        return self._data

    @property
    def covariance_matrix(self):                # optional
        Cdinv = self.rf.InvDataCov(2.5,0.01,len(self.data))
        Cdinv /= 100        # temper the likelihood by rescaling the data covariance
        return Cdinv

    @property
    def inverse_covariance_matrix(self):
        raise NotImplementedError               # optional

    def _model_setup(self, model):
        model_setup = np.copy(self._ref_model_setup)
        if self.example_number == 1:
            model_setup[self._interfaces,0] = model
        elif self.example_number == 2:
            model_setup[:,1] = model
        elif self.example_number == 3:
            model_setup[:,:2] = model.reshape(model_setup[:,:2].shape)
        return model_setup

    def forward(self, model, with_jacobian=False, *args, **kwargs):
        if with_jacobian:
            raise NotImplementedError           # optional
        else:
            model_setup = self._model_setup(model)
            _, rfunc = self.rf.rfcalc(model_setup, *args, **kwargs)
            return rfunc

    def jacobian(self, model):
        raise NotImplementedError               # optional

    def plot_model(self, model):
        model_setup = self._model_setup(model)
        px = np.zeros([2*len(model_setup),2])
        px[0::2,0] = model_setup[:,1]
        px[1::2,0] = model_setup[:,1]
        px[1::2,1] = model_setup[:,0]
        px[2::2,1] = model_setup[:-1,0]
        fig, ax = plt.subplots(1, 1, figsize=(4,6))
        ax.set_xlabel('Vs (km/s)')
        ax.set_ylabel('Depth (km)')
        ax.invert_yaxis()
        ax.plot(px[:,0],px[:,1],'b-')
        return fig
    
    def plot_data(self, data, data2=None, label=None, label2=None):
        fig, ax = plt.subplots(1, 1)
        ax.plot(self._t,data,label=label)
        if data2 is not None:
            ax.plot(self._t,data2,'r-',label=label2)
        ax.set_xlabel('Time/s')
        ax.set_ylabel('Amplitude')
        ax.grid(True)
        return fig

    def misfit(self, data, data2):
        raise NotImplementedError               # optional

    def log_likelihood(self,data,data2):
        Cdinv = self.covariance_matrix
        res = data2 - data
        logLike = -0.5*np.dot(res,np.transpose(np.dot(Cdinv, res)))/2.0
        return logLike.item()

    def log_prior(self, model):
        if self.example_number == 1:
            depths_in_0_60 = all([m_p < 60.0 and m_p > 0. for m_p in model])
            if depths_in_0_60: return np.log(1/60).item()
        elif self.example_number == 2:
            veloc_in_4_6 = all([m_p < 60.0 and m_p > 0. for m_p in model])
            if veloc_in_4_6: return np.log(1/2).item()
        elif self.example_number == 3:
            depths_in_0_60 = all([m_p < 60 and m_p > 0 for m_p in model[[0,2,4,6,8]]])
            veloc_in_4_6 = all([m_p < 60 and m_p > 0 for m_p in model[[1,3,5,7,9]]])
            if depths_in_0_60 and veloc_in_4_6:
                return np.log(1/60).item()
        return float("-inf")
